﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading;
using System.Xml;
using Microsoft.SqlServer.Management.Common;
using Microsoft.SqlServer.Management.Dac;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.Diagnostics;
using Microsoft.WindowsAzure.ServiceRuntime;
using Microsoft.WindowsAzure.StorageClient;
using System.Collections.Specialized;

namespace WorkerRole
{
    public class WorkerRole : RoleEntryPoint
    {
        private string inputXML = string.Empty;
        private LogDataContext logContext;

        public override void Run()
        {
            logContext = new LogDataContext();
            //inputXML = GetInputXML();
            //if (inputXML.Trim().Length > 0)
            //{
                while (true)
                {
                    inputXML = GetInputXML();
                    if (inputXML.Trim().Length > 0)
                    {
                        BackupSQLAzureDatabase();
                        Thread.Sleep(1000 * 60 * 5); // Every 5 min
                    }
                }
            //}
        }

        public string GetInputXML()
        {
            try
            {
                CloudStorageAccount storageAccount;
                CloudBlobClient blobClient;

                storageAccount = CloudStorageAccount.FromConfigurationSetting("InputXMLBlobConnectionString");

                blobClient = storageAccount.CreateCloudBlobClient();
                blobClient.RetryPolicy = RetryPolicies.Retry(4, TimeSpan.Zero);

                CloudBlobContainer blobContainer = blobClient.GetContainerReference(RoleEnvironment.GetConfigurationSettingValue("InputXMLBlobContainerName"));
                CloudBlob blob = blobContainer.GetBlobReference("SQLAzureBackupData.xml");

                return blob.DownloadText();
            }
            catch (Exception Ex)
            {
                logContext.SaveError("Getting Input XML cause issue. The error: " + Ex.Message + " at " + DateTime.Now.ToString("ddMMMMyyyy HHMMSS"));
            }
            return string.Empty;
        }

        public void PutInputXML(string xmlData)
        {
            try
            {
                CloudStorageAccount storageAccount;
                CloudBlobClient blobClient;

                storageAccount = CloudStorageAccount.FromConfigurationSetting("InputXMLBlobConnectionString");

                blobClient = storageAccount.CreateCloudBlobClient();
                blobClient.RetryPolicy = RetryPolicies.Retry(4, TimeSpan.Zero);

                CloudBlobContainer blobContainer = blobClient.GetContainerReference(RoleEnvironment.GetConfigurationSettingValue("InputXMLBlobContainerName"));
                CloudBlob blob = blobContainer.GetBlobReference("SQLAzureBackupData.xml");

                blob.UploadText(xmlData);
            }
            catch (Exception Ex)
            {
                logContext.SaveError("Getting Input XML cause issue. The error: " + Ex.Message + " at " + DateTime.Now.ToString("ddMMMMyyyy HHMMSS"));
            }
        }

        public override bool OnStart()
        {
            // Set the maximum number of concurrent connections 
            ServicePointManager.DefaultConnectionLimit = 12;

            // This code sets up a handler to update CloudStorageAccount instances when their corresponding
            // configuration settings change in the service configuration file.
            CloudStorageAccount.SetConfigurationSettingPublisher((configName, configSetter) =>
            {
                // Provide the configSetter with the initial value
                configSetter(RoleEnvironment.GetConfigurationSettingValue(configName));
            });

            // For information on handling configuration changes
            // see the MSDN topic at http://go.microsoft.com/fwlink/?LinkId=166357.
            RoleEnvironment.Changing += RoleEnvironmentChanging;

            return base.OnStart();
        }

        private void RoleEnvironmentChanging(object sender, RoleEnvironmentChangingEventArgs e)
        {
            if (e.Changes.Any(change => change is RoleEnvironmentConfigurationSettingChange))
            {
                e.Cancel = true;
            }
        }

        public bool BackupSQLAzureDatabase()
        {
            logContext.SaveEvent("Starting the Backup Process");
            //logContext.SaveEvent("XML = " + inputXML);

            if (inputXML.Trim().Length > 0)
            {

                XmlDocument doc = new XmlDocument();
                doc.LoadXml(inputXML);

                try
                {
                    foreach (XmlNode node in doc.DocumentElement.SelectNodes("Database"))
                    {
                        #region Get Values from XML
                        DatabaseInfo databaseInfo = new DatabaseInfo();

                        databaseInfo.ServerName = (node.Attributes["ServerName"] != null ? node.Attributes["ServerName"].Value : string.Empty);
                        databaseInfo.DatabaseName = (node.Attributes["DatabaseName"] != null ? node.Attributes["DatabaseName"].Value : string.Empty);
                        databaseInfo.UseWindowsAuth = (node.Attributes["UseWindowsAuth"] != null ? Convert.ToBoolean(node.Attributes["UseWindowsAuth"].Value) : false);
                        databaseInfo.LoginId = (node.Attributes["LoginId"] != null ? node.Attributes["LoginId"].Value : string.Empty);
                        databaseInfo.Password = (node.Attributes["Password"] != null ? node.Attributes["Password"].Value : string.Empty);
                        databaseInfo.BackupFileName = (node.Attributes["BackupFileName"] != null ? node.Attributes["BackupFileName"].Value.Replace("@Date", DateTime.Now.ToString("ddMMMMyyyy")).Replace("@Time", DateTime.Now.ToString("hhmmss")) : "Backup" + databaseInfo.DatabaseName + DateTime.Now.ToString("ddMMMyyyy hhmm"));
                        databaseInfo.BackupScheduledPeriod = (node.Attributes["BackupScheduledPeriod"] != null ? (BackupPeriod)Enum.Parse(typeof(BackupPeriod), node.Attributes["BackupScheduledPeriod"].Value) : BackupPeriod.Daily);
                        databaseInfo.BackupScheduledTime = (node.Attributes["BackupScheduledTime"] != null ? Convert.ToDateTime(node.Attributes["BackupScheduledTime"].Value) : DateTime.Today);
                        databaseInfo.LastBackupDateTime = (node.Attributes["LastBackupDateTime"].Value.Trim().Length == 0 ? System.DateTime.Now : Convert.ToDateTime(node.Attributes["LastBackupDateTime"].Value));
                        databaseInfo.ExportTo = (node.Attributes["ExportTo"] != null ? (ExportTo)Enum.Parse(typeof(ExportTo), node.Attributes["ExportTo"].Value) : ExportTo.FileSystem);
                        databaseInfo.ExportFileDirectory = (node.Attributes["ExportFileDirectory"] != null ? node.Attributes["ExportFileDirectory"].Value.Replace("@Date", DateTime.Now.ToString("ddMMMMyyyy")).Replace("@Time", DateTime.Now.ToString("hhmmss")) : string.Empty);
                        databaseInfo.ExportFileDirectory = (databaseInfo.ExportFileDirectory != string.Empty ? (databaseInfo.ExportFileDirectory.EndsWith(@"\") == true ? databaseInfo.ExportFileDirectory : databaseInfo.ExportFileDirectory + @"\") : @"C:\SQLAzureBackup");
                        databaseInfo.TrustServerCertificate = (node.Attributes["TrustServerCertificate"] != null ? Convert.ToBoolean(node.Attributes["TrustServerCertificate"].Value) : false);
                        databaseInfo.UseSSL = (node.Attributes["UseSSL"] != null ? Convert.ToBoolean(node.Attributes["UseSSL"].Value) : false);
                        databaseInfo.StorageAccountName = (node.Attributes["StorageAccountName"] != null ? node.Attributes["StorageAccountName"].Value : string.Empty);
                        databaseInfo.BlobContainerName = (node.Attributes["BlobContainerName"] != null ? node.Attributes["BlobContainerName"].Value : string.Empty);
                        databaseInfo.BlobDefaultEndpointsProtocol = (node.Attributes["BlobDefaultEndpointsProtocol"] != null ? node.Attributes["BlobDefaultEndpointsProtocol"].Value : string.Empty);
                        databaseInfo.StorageAccessKey = (node.Attributes["StorageAccessKey"] != null ? node.Attributes["StorageAccessKey"].Value : string.Empty);
                        #endregion

                        logContext.SaveEvent("Starting Process for " + databaseInfo.DatabaseName + "@" + databaseInfo.ServerName);

                        bool IsProceed = false;
                        logContext.SaveEvent("Validation \n LastBackupDateTime : " + databaseInfo.LastBackupDateTime.ToString("dd-MMM-yyyy") + " <= " + System.DateTime.Now.ToString("dd-MMM-yyyy") +
                                                        "\n BackupScheduledHours : " + databaseInfo.BackupScheduledTime.Hour.ToString() + " == " + System.DateTime.Now.Hour.ToString() +
                                                        "\n and Minutes: " + databaseInfo.BackupScheduledTime.Minute.ToString() + " <= " + System.DateTime.Now.Minute.ToString());
                        #region Validation
                        if (databaseInfo.BackupScheduledPeriod == BackupPeriod.Daily)
                        {
                            if ((Convert.ToDateTime(databaseInfo.LastBackupDateTime.ToString("dd-MMM-yyyy")) <= Convert.ToDateTime(System.DateTime.Now.ToString("dd-MMM-yyyy")).AddDays(-1)) &&
                                (databaseInfo.BackupScheduledTime.Hour == System.DateTime.Now.Hour) &&
                                (databaseInfo.BackupScheduledTime.Minute <= System.DateTime.Now.Minute)
                                )
                                IsProceed = true;
                        }
                        #endregion

                        if (IsProceed == true)
                        {
                            logContext.SaveEvent("Backup Started for database " + databaseInfo.DatabaseName + "@" + databaseInfo.ServerName + " at " + DateTime.Now.ToString("ddMMMMyyyy HHMMSS"));

                            #region Backup Process
                            switch (databaseInfo.ExportTo)
                            {
                                case ExportTo.FileSystem:
                                    BackupToFileSystem(databaseInfo);
                                    break;
                                case ExportTo.Blob:
                                    BackupToBlob(databaseInfo);
                                    break;
                            }
                            #endregion

                            logContext.SaveEvent("Backup Completed for database '" + databaseInfo.DatabaseName + "' on server '" + databaseInfo.ServerName + "' at " + DateTime.Now.ToString("ddMMMMyyyy HHMMSS"));
                        }
                        else
                        {
                            logContext.SaveEvent("Validation failed for database " + databaseInfo.DatabaseName + "@" + databaseInfo.ServerName);
                        }
                        node.Attributes["LastBackupDateTime"].Value = System.DateTime.Now.ToString("dd-MMM-yyyy hh:mm tt");
                        PutInputXML(doc.InnerXml);
                    }

                    return true;
                }
                catch (Exception ex)
                {
                    logContext.SaveError(String.Format("Error Occured : {0} at {1}", ex.Message, System.DateTime.Now.ToString("dd-MMM-yyyy hh:mm")), ex.Message + "\\" + ex.StackTrace);
                }
                return false;
            }
            else
            {
                logContext.SaveError("XML File does not found", "Checking File Exist");
            }
            logContext.SaveEvent("Ending the Backup Process");
            return false;
        }

        private bool BackupToBlob(DatabaseInfo databaseInfo)
        {
            try
            {
                ServerConnection connection = this.GetServerConnection(databaseInfo);
                DacStore dacStore = null;

                // Get a DacStore for the current connection
                dacStore = new DacStore(connection);

                string backupFileName = string.Empty;

                backupFileName = Path.Combine(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location) + @"\", databaseInfo.DatabaseName + "_" + System.DateTime.Now.ToString("dd-MMM-yyyy hhmmss") + ".bacpac");
                dacStore.Export(databaseInfo.DatabaseName, backupFileName);

                CloudBlobClient blobStorage;

                // read account configuration settings
                string strConnectionString = "DefaultEndpointsProtocol=" + databaseInfo.BlobDefaultEndpointsProtocol + ";AccountName=" + databaseInfo.StorageAccountName + ";AccountKey=" + databaseInfo.StorageAccessKey;
                var storageAccount = CloudStorageAccount.Parse(strConnectionString);

                // create blob container for backup data
                blobStorage = storageAccount.CreateCloudBlobClient();
                CloudBlobContainer container = blobStorage.GetContainerReference(databaseInfo.BlobContainerName);
                container.CreateIfNotExist();

                // configure container for private access
                var permissions = container.GetPermissions();
                permissions.PublicAccess = BlobContainerPublicAccessType.Off;
                container.SetPermissions(permissions);

                CloudBlockBlob blob = container.GetBlockBlobReference(databaseInfo.BackupFileName);

                blob.Metadata["Server"] = databaseInfo.ServerName;
                blob.Metadata["Database"] = databaseInfo.DatabaseName;
                blob.Metadata["FileName"] = databaseInfo.BackupFileName;
                blob.Metadata["BackupDate"] = DateTime.Now.ToString("dd-MMM-yyyy HH:MM");

                blob.UploadFile(backupFileName);
                System.IO.File.Delete(backupFileName);

                connection.Disconnect();

                return true;
            }
            catch (BacpacException bacpacex)
            {
                logContext.SaveError("Error (BacpacException) when backup", bacpacex.Message + "\n" + bacpacex.StackTrace);
            }
            catch (DacException dacex)
            {
                logContext.SaveError("Error (DacException) when backup", dacex.Message + "\n" + dacex.StackTrace);
            }
            catch (Exception ex)
            {
                logContext.SaveError("Error when backup", ex.Message + "\n" + ex.StackTrace);
            }
            finally
            {
            }
            return false;
        }

        private bool BackupToFileSystem(DatabaseInfo databaseInfo)
        {
            try
            {
                ServerConnection connection = this.GetServerConnection(databaseInfo);
                DacStore dacStore = null;

                // Get a DacStore for the current connection
                dacStore = new DacStore(connection);

                string backupFileName = string.Empty;

                backupFileName = Path.Combine(databaseInfo.ExportFileDirectory, databaseInfo.BackupFileName);
                if (!Directory.Exists(databaseInfo.ExportFileDirectory))
                    Directory.CreateDirectory(databaseInfo.ExportFileDirectory);

                dacStore.Export(databaseInfo.DatabaseName, backupFileName);

                connection.Disconnect();

                return true;
            }
            catch (BacpacException bacpacex)
            {
                logContext.SaveError("Error (BacpacException) when backup", bacpacex.Message + "\n" + bacpacex.StackTrace);
            }
            catch (DacException dacex)
            {
                logContext.SaveError("Error (DacException) when backup", dacex.Message + "\n" + dacex.StackTrace);
            }
            catch (Exception ex)
            {
                logContext.SaveError("Error when backup", ex.Message + "\n" + ex.StackTrace);
            }
            finally
            {
            }
            return false;
        }

        private ServerConnection GetServerConnection(DatabaseInfo databaseInfo)
        {
            SqlConnectionStringBuilder sqlConnectionStringBuilder = new SqlConnectionStringBuilder();

            sqlConnectionStringBuilder.DataSource = databaseInfo.ServerName;

            sqlConnectionStringBuilder.TrustServerCertificate = databaseInfo.TrustServerCertificate;

            if (!string.IsNullOrEmpty(databaseInfo.DatabaseName))
                sqlConnectionStringBuilder.InitialCatalog = databaseInfo.DatabaseName;

            sqlConnectionStringBuilder.Encrypt = databaseInfo.UseSSL;

            if (databaseInfo.UseWindowsAuth == false)
            {
                if (databaseInfo.LoginId == null)
                {
                    throw new ArgumentException("User name must be specified unless Windows Authentication (-E) is specified.");
                }
                if (databaseInfo.Password == null)
                {
                    throw new InvalidArgumentException("User password must be specified unless Windows Authentication (-E) is specified.");
                }
                sqlConnectionStringBuilder.IntegratedSecurity = false;
                sqlConnectionStringBuilder.UserID = databaseInfo.LoginId;
                sqlConnectionStringBuilder.Password = databaseInfo.Password;
            }
            else
            { // To do
                sqlConnectionStringBuilder.IntegratedSecurity = true;
            }

            sqlConnectionStringBuilder.ConnectTimeout = 60;
            sqlConnectionStringBuilder.MinPoolSize = 5;
            sqlConnectionStringBuilder.MaxPoolSize = 50;        // According to Azure team there is NO max limit on connections.  It all comes down to CPU and IO usage.  They said setting this to 100 would be fine.
            sqlConnectionStringBuilder.Pooling = true;

            // To do
            //// If we think we are talking to Azure we should turn on encryption if the user didn't set it
            //if ((this.azureEdition != AzureEdition.Default || this.azureSize != -1) && this.useSSL == false)
            //{
            //    sqlConnectionStringBuilder.Encrypt = true;
            //}

            ServerConnection connection = new ServerConnection();
            connection.ConnectionString = sqlConnectionStringBuilder.ToString();

            connection.StatementTimeout = 60 * 60 * 24;
            connection.Connect();
            return connection;
        }
    }
}
